"""
 - 提供主要的训练流
    - 加载数据集 DataLoader
    - train、test、eval
 - 提供对训练流的控制、回调
    - callbacks
 - 提供对训练中间的状态保存
    - metric: Meter
    - checkpoints: Saver


trainer = Trainer()
"""
import inspect
import os
import sys
from functools import wraps
from typing import Callable

import numpy as np
import six
import torch
from torch.optim.optimizer import Optimizer

from .components import TrainerPropVar
from .. import TrainStage


def _exit_hook_v0(exc_type, exc, tb, *args):
    import traceback
    res = traceback.format_exception(exc_type, exc, tb)
    res = [i for i in res if 'in _newfunc' not in i]
    print(''.join(res), file=sys.stderr)


def _exit_hook(exc_type, exc, tb, *args):
    from rich.console import Console
    console = Console()
    _exit_hook_v0(exc_type, exc, tb, *args)
    try:
        six.reraise(exc_type, exc, tb)
    except:
        from rich.traceback import Traceback
        traceback = Traceback(
            width=100,
            extra_lines=3,
            theme=None,
            word_wrap=False,
            show_locals=False,
            suppress=(),
            max_frames=100,
        )
        for stack in traceback.trace.stacks:
            stack.frames = [i for i in stack.frames if all([i.name != k for k in {'_newfunc'}])]
        console.print(traceback)


def wrapper(self, func, _call_set: list):
    """
    对每个 Trainer 的 _call_backs 类变量中定义的函数尝试绑定回调
    Args:
        func:
        _call_set:

    Returns:

    """

    @wraps(func)
    def _newfunc(*aargs, **kkwargs):
        """执行前回调 on_begin() 、执行后回调 on_end()、执行异常则回调 on_exception() """
        for callback in _call_set:
            callback.on_begin(self, func, self.params, *aargs, **kkwargs)
        try:
            _meter = func(*aargs, **kkwargs)
        except BaseException as e:
            _handles = [callback.on_exception(self, func, self.params, e, *aargs, **kkwargs)
                        for callback in _call_set]

            if any(_handles):
                return None
            else:
                raise e

        for callback in _call_set:
            callback.on_end(self, func, self.params, _meter, *aargs, **kkwargs)
        return _meter

    return _newfunc


init_function = ['icallbacks', 'imodels']
call_dependency = {
    'train': init_function,
}


class _BaseTrainer:
    __exp_name__ = None

    callback_function = {}

    def __new__(cls, *args, **kwargs):
        self = super().__new__(cls)
        if cls.__exp_name__ is None:
            cls.__exp_name__ = cls.__name__.lower().replace("trainer", "Exp")

        self._prop = {}
        self._cmp = {}
        self._rev_index = {}
        self._call_order = {}

        self._state_dicts = {
            'optims': set(),
            'models': set(),
            'others': set(),
            'tensor.th': set(),
            'tensor.np': set(),
        }

        from lumo.utils.exithook import replace

        replace(_exit_hook)

        def init_wrapper(func):
            @wraps(func)
            def inner(dm=None, params=None, *args, **kwargs):
                init_fn = getattr(self, 'initialize', None)
                if init_fn is not None:
                    init_fn()
                process_loader = getattr(self, 'process_loader', None)
                if process_loader is not None:
                    process_loader(dm, TrainStage.create_from_str(func.__name__))
                func(*args, **kwargs)

            return inner

        def cb_wrapper(func, call_set: list):
            """
            对每个 Trainer 的 _call_backs 类变量中定义的函数尝试绑定回调
            Args:
                func:
                call_set:

            Returns:

            """

            @wraps(func)
            def _newfunc(*aargs, **kkwargs):
                """执行前回调 on_begin() 、执行后回调 on_end()、执行异常则回调 on_exception() """
                # on_begin
                for callback in call_set:
                    callback.on_begin(self, func, self.params, *aargs, **kkwargs)
                try:
                    _meter = func(*aargs, **kkwargs)
                except BaseException as e:
                    _handles = [callback.on_exception(self, func, self.params, e, *aargs, **kkwargs)
                                for callback in call_set]
                    self.on_trainer_exception(func, e)
                    if any(_handles):
                        return None
                    else:
                        raise e

                for callback in call_set:
                    callback.on_end(self, func, self.params, _meter, *aargs, **kkwargs)
                return _meter

            return _newfunc

        self.callbacks = []

        vars = dir(self)
        for name in vars:
            if name not in self.callback_function:
                continue
            value = getattr(self, name, None)
            if value is None:
                continue
            if callable(value):
                setattr(self, name, cb_wrapper(value, self.callbacks))

        for func in [self.train, self.test, self.evaluate]:
            setattr(self, func.__name__, init_wrapper(func))
        return self

    def __getitem__(self, item):
        return getattr(self, item)

    def __setitem__(self, key: str, value):
        self.__setattr__(key, value)

    def __setattr__(self, name, value):
        super().__setattr__(name, value)

        if name.startswith('_') or name.endswith('_'):
            # when callback is trainer itself, _hooked will be passed, and caused recursion
            # if some pretrained models need to be ignored in save/load stage, it can be named like 'some_' or '_some'
            return

        if isinstance(value, torch.device):
            type_name = 'devices'
        elif isinstance(value, torch.nn.Module):
            type_name = 'models'
        elif isinstance(value, Optimizer):
            type_name = 'optims'
        elif isinstance(value, torch.Tensor):
            type_name = 'tensor.th'
        elif isinstance(value, np.ndarray):
            type_name = 'tensor.np'
        elif callable(getattr(value, "state_dict", None)) and callable(getattr(value, "load_state_dict", None)):
            type_name = 'others'
        else:
            super().__setattr__(name, value)
            return

        # if name in self.__dict__: TODO workaround multi-gpu error: Expected to mark a variable ready only once
        #     self.__dict__.pop(name)

        self._state_dicts.setdefault(type_name, set()).add(name)
        self._rev_index[name] = type_name

    # def __getattr__(self, name):
    #     if name.startswith('_') or name.endswith('_'):
    #         # when callback is trainer itself, _hooked will be passed, and caused recursion
    #         # if some pretrained models need to be ignored in save/load stage, it can be named like 'some_' or '_some'
    #         raise AttributeError(name)
    #     type_name = self._rev_index.get(name, None)
    #     if type_name is None:
    #         raise AttributeError(name)
    #
    #     return self._state_dicts[type_name][name]

    @classmethod
    def dirname(cls):
        file = inspect.getfile(cls)
        return os.path.basename(os.path.dirname(file))

    @classmethod
    def filebasename(cls):
        try:
            file = inspect.getfile(cls)
            pre = os.path.splitext(os.path.basename(file))[0]
        except:
            pre = 'builtin'
        return pre

    @classmethod
    def generate_exp_name(cls) -> str:
        pre = cls.filebasename()

        exp_name = cls.__exp_name__
        if exp_name is None:
            exp_name = cls.__name__.lower().replace("trainer", "exp")

        return "{}.{}".format(pre.lower(), exp_name.lower())

    def on_trainer_exception(self, func: Callable, exception: BaseException):
        pass



import importlib

from accelerate import __version__ as accelerate_version
from fire import __version__ as fire_version
from joblib import __version__ as joblib_version
from psutil import __version__ as psutil_version

from lumo import __version__ as lumo_version


class Version:
    # lumo = lumo_version
    joblib = joblib_version
    fire = fire_version
    psutil = psutil_version
    accelerate = accelerate_version


def get_lock(*others):
    res = {}
    res['lumo'] = lumo_version
    res.update({k: v for k, v in Version.__dict__.items() if not k.startswith('__')})

    for lib in others:
        mod = importlib.import_module(lib)
        res[lib] = getattr(mod, '__version__', 'null')

    return res



"""
Safe means Exceptions won't be raised. When
 - call `dump_xxx` methods, True/False will be returned to indicate success/failed.
 - call `load_xxx` methods, None will be returned if something wrong happened.
"""
import json
import os
import pickle as _pickle
from contextlib import contextmanager
from io import FileIO

import torch
from joblib.numpy_pickle import dump as dump_nd, load as load_nd

dump_nd = dump_nd
load_nd = load_nd


def dump_json(obj, fn):
    with open(fn, 'w', encoding='utf-8') as w:
        json.dump(obj, w, indent=2)


def dump_yaml(obj, fn):
    import yaml
    with open(fn, 'w', encoding='utf-8') as w:
        yaml.safe_dump(obj, w)
    return fn


def dump_state_dict(obj, fn):
    torch.save(obj, fn)
    return fn


def load_json(fn):
    with open(fn, 'r', encoding='utf-8') as r:
        return json.load(r)


def load_yaml(fn):
    import yaml
    with open(fn, 'r', encoding='utf-8') as r:
        return yaml.safe_load(r)


def load_state_dict(fn: str, map_location='cpu'):
    ckpt = torch.load(fn, map_location=map_location)
    return ckpt


def load_text(fn):
    if not os.path.exists(fn):
        return ''
    with open(fn, 'r', encoding='utf-8') as r:
        return ''.join(r.readlines())


def dump_text(string: str, fn, append=False):
    mode = 'w'
    if append:
        mode = 'a'
    with open(fn, mode, encoding='utf-8') as w:
        w.write(string)
    return fn


def safe_getattr(self, key, default=None):
    try:
        return getattr(self, key, default)
    except:
        return default


def dump_pkl(obj, file, make_path=True, protocol=None, *, fix_imports=True):
    if isinstance(file, str):
        if make_path:
            os.makedirs(os.path.dirname(os.path.abspath(file)), exist_ok=True)
        file = open(file, 'wb')
        _pickle.dump(obj, file, protocol=protocol, fix_imports=fix_imports)
        file.close()
    elif isinstance(file, FileIO):
        _pickle.dump(obj, file, protocol=protocol, fix_imports=fix_imports)
    else:
        raise NotImplementedError()


def load_pkl(file, *, fix_imports=True, encoding="ASCII", errors="strict"):
    if isinstance(file, str):
        file = open(file, 'rb')
        res = _pickle.load(file, fix_imports=fix_imports, encoding=encoding, errors=errors)
        file.close()
        return res
    elif isinstance(file, FileIO):
        return _pickle.load(file, fix_imports=fix_imports, encoding=encoding, errors=errors)
    else:
        raise NotImplementedError()


@contextmanager
def cached(fn):
    import shutil
    cache_fn = f'{fn}.lumo_cache'
    try:
        yield cache_fn
    except:
        os.remove(cache_fn)
    finally:
        shutil.move(cache_fn, fn)



import torch
from torch.nn import functional as F

import torch
from torch.nn import functional as F


def normalize(feature: torch.Tensor, inplace=False, eps=1e-08):
    """
    a = torch.rand(2, 3)
    b = torch.rand(2, 3)
    sim_1 = batch_cosine_similarity2(a, b)
    sim_2 = batch_cosine_similarity(a, b)
    print(a)
    print(b)
    print(sim_1)
    print(sim_2)
    print((sim_1 - sim_2) ** 2)
    Args:
        feature:
        inplace:
        eps:

    Returns:

    """
    if inplace:
        feature /= (feature.norm(dim=-1, keepdim=True) + eps)
    else:
        feature = feature / (feature.norm(dim=-1, keepdim=True) + eps)
    return feature


def batch_cosine_similarity(a: torch.Tensor, b: torch.Tensor, eps=1e-08):
    """
    a: [bs_a, feature_dim]
    a: [bs_b, feature_dim]
    """
    return F.cosine_similarity(a.unsqueeze(1), b.unsqueeze(0), dim=2, eps=eps)


def batch_cosine_similarity2(a: torch.Tensor, b: torch.Tensor, eps=0):
    """
    a: [bs_a, feature_dim]
    a: [bs_b, feature_dim]
    """

    return torch.mm(normalize(a, eps=eps), normalize(b, eps=eps).T)


def masked_log_softmax(logits, mask=None, dim=-1):
    """
    Examples:
        >>> logits = torch.rand(4,10)
        >>> log_prob = masked_log_softmax(logits)
        >>> ((logits.masked_log_softmax(dim=-1) - log_prob)**2 < 1e-07).all()

    Args:
        logits:

    Returns:

    """
    logits = logits - logits.max(dim=dim, keepdim=True).values  # avoid nan and inf when logits value is too large
    exp_logits = torch.exp(logits)
    if mask is not None:
        exp_logits = exp_logits * mask.float()
    return logits - torch.log(exp_logits.sum(dim=dim, keepdim=True))


def masked_softmax(logits, mask=None, dim=-1, eps=1e-08):
    """
    see https://discuss.pytorch.org/t/apply-mask-softmax/14212/7
    Args:
        logits:
        mask:
        dim:
        eps:

    Returns:

    """
    if mask is None:
        return torch.softmax(logits, dim=dim)
    else:
        # matrix A is the one you want to do mask softmax at dim=1
        A_exp = torch.exp(logits - torch.max(logits, dim=1, keepdim=True).values)
        A_exp = A_exp * mask.float()  # this step masks
        res = A_exp / (torch.sum(A_exp, dim=1, keepdim=True) + eps)
        return res


def normalize(feature: torch.Tensor, inplace=False, dim=-1, eps=1e-08):
    if inplace:
        feature /= (feature.norm(dim=dim, keepdim=True) + eps)
    else:
        feature = feature / (feature.norm(dim=dim, keepdim=True) + eps)
    return feature


def batch_cosine_similarity(a: torch.Tensor, b: torch.Tensor, eps=1e-08):
    """
    a: [bs_a, feature_dim]
    a: [bs_b, feature_dim]
    """
    return F.cosine_similarity(a.unsqueeze(1), b.unsqueeze(0), dim=2, eps=eps)


def batch_cosine_similarity2(a: torch.Tensor, b: torch.Tensor, eps=0):
    """
    a: [bs_a, feature_dim]
    a: [bs_b, feature_dim]
    """

    return torch.mm(normalize(a, eps=eps), normalize(b, eps=eps).T)


def sharpen(x: torch.Tensor, t=1):
    """
    让概率分布变的更 sharp，即倾向于 onehot
    :param x: prediction, sum(x,dim=-1) = 1
    :param t: temperature, default is 0.5
    :return:
    """
    with torch.no_grad():
        temp = torch.pow(x, 1 / t)
        return temp / (temp.sum(dim=1, keepdim=True) + 1e-7)


if __name__ == '__main__':
    a = torch.rand(2, 3)
    b = torch.rand(2, 3)
    sim_1 = batch_cosine_similarity2(a, b)
    sim_2 = batch_cosine_similarity(a, b)
    print(a)
    print(b)
    print(sim_1)
    print(sim_2)
    print((sim_1 - sim_2) ** 2)



"""
https://stackoverflow.com/questions/8230638/parallel-coordinates-plot-in-matplotlib
"""
import matplotlib.pyplot as plt
from matplotlib.path import Path
import matplotlib.patches as patches
import numpy as np
from sklearn import datasets

iris = datasets.load_iris()
ynames = iris.feature_names
ys = iris.data[:20]
ymins = ys.min(axis=0)
ymaxs = ys.max(axis=0)

dys = ymaxs - ymins
ymins -= dys * 0.05  # add 5% padding below and above
ymaxs += dys * 0.05

ymaxs[1], ymins[1] = ymins[1], ymaxs[1]  # reverse axis 1 to have less crossings
dys = ymaxs - ymins

# transform all data to be compatible with the main axis
zs = np.zeros_like(ys)
zs[:, 0] = ys[:, 0]
zs[:, 1:] = (ys[:, 1:] - ymins[1:]) / dys[1:] * dys[0] + ymins[0]

fig, host = plt.subplots(figsize=(10,4))

axes = [host] + [host.twinx() for i in range(ys.shape[1] - 1)]
for i, ax in enumerate(axes):
    ax.set_ylim(ymins[i], ymaxs[i])
    ax.spines['top'].set_visible(False)
    ax.spines['bottom'].set_visible(False)
    if ax != host:
#         ax.spines['left'].set_visible(False)
        ax.yaxis.set_ticks_position('right')
        ax.spines["right"].set_position(("axes", i / (ys.shape[1] - 1)))

host.set_xlim(0, ys.shape[1] - 1)
host.set_xticks(range(ys.shape[1]))
host.set_xticklabels(ynames, fontsize=14)
host.tick_params(axis='x', which='major', pad=7)
host.spines['right'].set_visible(False)
host.xaxis.tick_top()
host.set_title('Parallel Coordinates Plot — Iris', fontsize=18, pad=12)

colors = plt.cm.Set2.colors
legend_handles = [None for _ in iris.target_names]
for j in range(ys.shape[0]):
    # create bezier curves
    verts = list(zip([x for x in np.linspace(0, len(ys) - 1, len(ys) * 3 - 2, endpoint=True)],
                     np.repeat(zs[j, :], 3)[1:-1]))
    codes = [Path.MOVETO] + [Path.CURVE4 for _ in range(len(verts) - 1)]
    path = Path(verts, codes)
    patch = patches.PathPatch(path, facecolor='none', lw=2, alpha=0.7, edgecolor=colors[iris.target[j]])
    legend_handles[iris.target[j]] = patch
    host.add_patch(patch)
host.legend(legend_handles, iris.target_names,
            loc='lower center', bbox_to_anchor=(0.5, -0.18),
            ncol=len(iris.target_names), fancybox=True, shadow=True)
plt.tight_layout()
plt.show()


"""

"""
import queue
from collections import defaultdict


class tree(dict):
    """Implementation of perl's autovivification feature."""

    def __getitem__(self, item):
        try:
            return dict.__getitem__(self, item)
        except KeyError:
            value = self[item] = type(self)()
            return value

    def walk(self):
        for k, v in self.items():
            yield k, v
            if isinstance(v, tree):
                for kk, vv in v.walk():
                    yield f'{k}/{kk}', vv


class Node:
    HEAD = 0
    MID = 1
    TAIL = 2

    def __init__(self):
        self.value = None
        self.link = []
        self.stage = None

    @property
    def is_head(self):
        return self.stage == self.HEAD

    @property
    def is_mid(self):
        return self.stage == self.MID

    @property
    def is_tail(self):
        return self.stage == self.TAIL

    def set_stage(self, stage):
        self.stage = stage
        return self

    def set_value(self, val):
        self.value = val
        return self

    def add_link(self, y):
        self.link.append(y)

    def __repr__(self):
        return f'Node({self.stage} ,{len(self.link)}, {self.value})'


class Forest:
    def __init__(self):
        self.dic = defaultdict(Node)
        self.order = []
        self.tail = set()

    def add_head(self, x, val=None):
        self.dic[x].set_value(val).set_stage(Node.HEAD)
        self.order.append(x)
        return self

    def check_node_type(self, x):
        return x in self.dic

    def add_link(self, x, y, y_val=None):
        assert x in self.dic, f'x must already existed in graph, has {self.order}, got {x}'
        assert y not in self.dic, f'y must be a new node in graph, has {self.order}, got {y}'
        self.dic[x].add_link(y)
        self.dic[y].set_value(y_val).set_stage(Node.MID)
        self.order.append(y)
        return self

    def add_tail(self, x, y, y_val=None):
        assert x in self.dic, f'x must already existed in graph, has {self.order}, got {x}'
        assert y not in self.dic, f'y must be a new node in graph, has {self.order}, got {y}'
        self.dic[x].add_link(y)
        self.dic[y].set_value(y_val).set_stage(Node.TAIL)
        self.order.append(y)
        self.tail.add(y)
        return self

    def __iter__(self):
        stack = []
        mem = set()

        if len(self.order) > 0:
            stack.append(self.order[0])

        while len(stack) > 0:
            key = stack.pop(0)
            if key in mem:
                continue
            yield key, self.dic[key]

            mem.add(key)
            for key in self.dic[key].link:
                stack.append(key)


if __name__ == '__main__':
    dag = Forest()

    dag.add_head('1', 1)
    dag.add_link('1', '2', 2)
    dag.add_tail('2', '3')
    dag.add_tail('1', '4', 13)
    for i in dag:
        print(i)



from functools import wraps
from importlib import util as imputil

_lib_memory = set()


def is_lib_available(lib):
    if lib in _lib_memory:
        return True
    res = imputil.find_spec(lib)
    if res is None:
        return False

    _lib_memory.add(lib)
    return True


def torch_required(func):
    # Chose a different decorator name than in tests so it's clear they are not the same.
    @wraps(func)
    def wrapper(*args, **kwargs):
        if is_lib_available('torch'):
            return func(*args, **kwargs)
        else:
            raise ImportError(f"Method `{func.__name__}` requires PyTorch.")

    return wrapper


def lib_required(lib_name):
    def outer(func):
        @wraps(func)
        def wrapper(*args, **kwargs):
            if is_lib_available(lib_name):
                return func(*args, **kwargs)
            else:
                raise ImportError(f"Method `{func.__name__}` requires {lib_name}.")

        return wrapper

    return outer



import os
from itertools import chain
from typing import Dict, List

try:
    import regex as re
except ImportError:
    import re

from pathlib import Path
from collections import defaultdict
from dataclasses import dataclass
from numbers import Number
from lumo.exp import Experiment, finder

match_metric_v0 = re.compile(' ([a-z0-9A-Z_]+): ([0-9.]+)')
match_metric_v1 = re.compile(' ([a-z0-9A-Z_])=([0-9.]+)')


@dataclass()
class Step:
    value: Number
    step: int


def find_metric_fron_test_root(test_root):
    test_root = finder.ensure_test_root(test_root)
    if test_root is None:
        return False, {}

    exp = Experiment.from_disk(test_root)
    if exp.has_prop('tensorboard_args'):
        tb = exp.get_prop('tensorboard_args')
        metrics = parse_fron_tensorboard(tb['log_dir'])
    elif exp.has_prop('logger_args'):
        tb = exp.get_prop('logger_args')
        metrics = parse_from_log(tb['log_dir'])
    else:
        fs = [i for i in os.listdir(exp.test_root)]
        if len([f for f in fs if f.endswith('.log')]) > 0:
            metrics = parse_from_log(os.path.join(exp.test_root, fs[0]))
        else:
            metrics = {}
    return True, metrics


def parse_from_log(log) -> Dict[str, List[Step]]:
    metrics = defaultdict(list)
    with open(log) as r:
        for line in r:
            temp1 = match_metric_v0.findall(line.strip())
            temp2 = match_metric_v0.findall(line.strip())
            for k, v in chain(temp1, temp2):
                try:
                    metrics[k].append(Step(value=float(v), step=len(metrics[k])))
                except ValueError:
                    pass
    return metrics


def parse_fron_tensorboard(root) -> Dict[str, List[Step]]:
    from lumo.vis.parser_tb import TBReader
    reader = TBReader(root)
    metrics = defaultdict(list)
    for k in reader.scalar_names:
        metrics[k] = [Step(i.value, i.step) for i in reader.Scalars(k)]
    return metrics



import os.path

import git
from git import RemoteProgress
from urllib3.util import parse_url
from lumo import Logger

log = Logger()
template_map = {
    'classify': 'https://github.com/pytorch-lumo/wsl-baselines'
}


def prograss(*args, **kwargs):
    log.inline(*args)


def git_clone(url, alias=None):
    if alias is None:
        u = parse_url(url)
        alias = u.path.split('/')[-1]

    res = git.Repo.clone_from(url, alias, progress=prograss)
    log.newline()
    return res, alias


def git_clone_from_template(template, alias=None):
    url = template_map[template]
    return git_clone(url, alias)



from operator import and_, gt, ge, le, lt, eq, ne
from itertools import accumulate


def in_(ser, value):
    return ser.apply(lambda x: x in value)


def not_in_(ser, value):
    return ser.apply(lambda x: x not in value)


def first(ser, value):
    return ser.duplicated(value) == False


mapping = {
    '>=': ge,
    '<=': le,
    '==': eq,
    '!=': ne,
    '>': gt,
    '<': lt,
    'in': in_,
    'notin': not_in_,
    'first': first,
}


class Compare:
    def __init__(self, name=None):
        self.name = name
        self.op = None
        self.value = None
        self.drop = False

    def __getattr__(self, item):
        return Compare(item)

    def __getitem__(self, item):
        return Compare(item)

    def __invert__(self):
        self.drop = ~self.drop
        return self

    def __neg__(self):
        self.drop = True
        return self

    def __ge__(self, other):
        assert other is not None
        self.value = other
        self.op = ">="
        return self

    def __le__(self, other):
        assert other is not None
        self.value = other
        self.op = "<="
        return self

    def __eq__(self, other):
        self.value = other
        self.op = "=="
        return self

    def __ne__(self, other):
        self.value = other
        self.op = "!="
        return self

    def __gt__(self, other):
        assert other is not None
        self.value = other
        self.op = ">"
        return self

    def __lt__(self, other):
        assert other is not None
        self.value = other
        self.op = "<"
        return self

    def __repr__(self):
        return f'{self.name} {self.op} {self.value}'

    def in_(self, lis):
        self.op = 'in'
        self.value = set(lis)
        return self

    def not_in_(self, lis):
        self.op = 'notin'
        self.value = set(lis)
        return self

    def first(self, value):
        self.op = 'first'
        self.value = value
        return self


C = Compare()


def filter_by_condition(df, *condition: Compare):
    filters = []
    remains = []
    drops = []
    for cond in condition:
        if cond.op is None:
            if cond.drop:
                drops.append(cond.name)
            else:
                remains.append(cond.name)
        else:
            if cond.value is None:
                if cond.op == '==':
                    res = df[cond.name].isna()
                else:
                    res = ~df[cond.name].isna()
            else:
                op = mapping[cond.op]
                res = op(df[cond.name], cond.value)
            filters.append(res)

    df = df[list(accumulate(filters, and_))[-1]]

    if len(remains) > 0:
        df = df[remains]
    if len(drops) > 0:
        df = df.drop(drops, axis=1)

    return df



"""
easydict, support dic['a.b'] to refer dic.a.b
"""
from collections import OrderedDict
from typing import List


class Attr(OrderedDict):

    def __setattr__(self, key: str, value):
        _set_item(self, key.split('.'), value)

    def __setitem__(self, key, value):
        if not isinstance(key, str):
            raise TypeError('Key in attr must be str')
        _set_item(self, key.split('.'), value)

    def __getattr__(self, key: str):
        try:
            res = _get_item(self, key.split('.'))
        except KeyError:
            res = Attr()
            _set_item(self, key.split('.'), res)

        return res

    def __getitem__(self, key):
        if not isinstance(key, str):
            raise TypeError('Key in attr must be str')
        return _get_item(self, key.split('.'))


def _set_item(dic, keys: List[str], value):
    if len(keys) == 1:
        if isinstance(value, dict):
            value = dic(value)
        OrderedDict.__setitem__(dic, keys[0], value)
    else:
        nex = Attr()
        OrderedDict.__setitem__(dic, keys[0], nex)
        _set_item(nex, keys[1:], value)


def _get_item(dic, keys: List[str]):
    if len(keys) == 1:
        return OrderedDict.__getitem__(dic, keys[0])
    else:
        nex = OrderedDict.__getitem__(dic, keys[0])
        if isinstance(nex, dict):
            return _get_item(nex, keys[1:])
        else:
            raise KeyError(keys)



import importlib
from typing import overload

from torch.optim import Optimizer

from lumo.core import interp
from lumo.core.params import BaseParams


class InterpFactory:
    Cos = interp.Cos
    Linear = interp.Linear
    Exp = interp.Exp
    Log = interp.Log
    Constant = interp.Constant
    PeriodCos = interp.PeriodCos
    PeriodHalfCos = interp.PeriodHalfCos
    PeriodTriangle = interp.PeriodTriangle
    PeriodLinear = interp.PeriodLinear
    PowerDecay = interp.PowerDecay
    List = interp.InterpolateList


class OptimBuilder(BaseParams):

    @classmethod
    def from_kwargs(cls, **kwargs):
        self = cls()
        self.update(kwargs)
        return self

    def build(self, parameters, optim_cls=None) -> Optimizer:
        res = self.copy()
        name = res['name']
        lname = name.lower()
        args = res.copy()
        args.pop('name')

        if optim_cls is None:
            if lname in {'lars'}:
                optim_lib = importlib.import_module("lumo.contrib.optim.{}".format(lname))
            else:
                optim_lib = importlib.import_module("torch.optim.{}".format(lname))
            optim_cls = getattr(optim_lib, name, None)

        return optim_cls(parameters, **args)


class _OptimFactory:
    @overload
    def create_optim(self, name='SGD', lr=None, momentum=0, dampening=0, weight_decay=0,
                     nesterov=False) -> OptimBuilder:
        pass

    @overload
    def create_optim(self, name='Adam', lr=1e-3, betas=(0.9, 0.999), eps=1e-8, weight_decay=0,
                     amsgrad=False) -> OptimBuilder:
        pass

    @overload
    def create_optim(self, name='Adadelta', lr=1.0, rho=0.9, eps=1e-6, weight_decay=0) -> OptimBuilder:
        pass

    @overload
    def create_optim(self, name='Adagrad', lr=1e-2, lr_decay=0, weight_decay=0, initial_accumulator_value=0,
                     eps=1e-10) -> OptimBuilder:
        pass

    @overload
    def create_optim(self, name='AdamW', lr=2e-3, betas=(0.9, 0.999), eps=1e-8, weight_decay=0) -> OptimBuilder:
        pass

    @overload
    def create_optim(self, name='AdamW',
                     lr=1e-3, betas=(0.9, 0.999), eps=1e-8, weight_decay=1e-2, amsgrad=False) -> OptimBuilder:
        pass

    @overload
    def create_optim(self, name='ASGD',
                     lr=1e-2, lambd=1e-4, alpha=0.75, t0=1e6, weight_decay=0) -> OptimBuilder:
        pass

    @overload
    def create_optim(self, name='LBFGS',
                     lr=1, max_iter=20, max_eval=None, tolerance_grad=1e-7, tolerance_change=1e-9, history_size=100,
                     line_search_fn=None) -> OptimBuilder:
        pass

    @overload
    def create_optim(self, name='RMSprop', lr=1e-2, alpha=0.99, eps=1e-8, weight_decay=0, momentum=0,
                     centered=False) -> OptimBuilder:
        pass

    @overload
    def create_optim(self, name='Rprop', lr=1e-2, etas=(0.5, 1.2), step_sizes=(1e-6, 50)) -> OptimBuilder:
        pass

    @overload
    def create_optim(self, name='SparseAdam', lr=1e-3, betas=(0.9, 0.999), eps=1e-8) -> OptimBuilder:
        pass

    def create_optim(self, name=None, **kwargs) -> OptimBuilder:
        return OptimBuilder.from_kwargs(name=name, **kwargs)


OptimFactory = _OptimFactory()



import numpy as np
from matplotlib import pyplot as plt
from lumo.contrib.itertools import window


def make_grid_from_files(files, pad=1, ncol=None, padvalue=0):
    return make_grid_from_nplist([plt.imread(f) for f in files], pad, ncol, padvalue)


def make_grid_from_nplist(vals, pad=1, ncol=None, padvalue=0):
    return make_grid_from_stacked_imgs(np.stack(vals), pad, ncol, padvalue)


def make_grid_from_stacked_imgs(imgs: np.ndarray, pad=1, ncol=None, padvalue=0):
    x = imgs.shape[0]

    ws = ncol
    if ws is None:
        ws = min(int(np.ceil(np.sqrt(x))), 8)
    else:
        ws = min(x, ws)
    hs = x // ws + int((x % ws) != 0)

    h, w = imgs.shape[1], imgs.shape[2]
    nh = hs * (h + pad) + pad
    nw = ws * (w + pad) + pad
    if len(imgs.shape) == 3:
        nimgs = np.zeros((nh, nw), dtype=imgs.dtype) + padvalue
    else:
        nimgs = np.zeros((nh, nw, imgs.shape[-1]), dtype=imgs.dtype) + padvalue

    for row, sub_imgs in enumerate(window(imgs, ws, ws)):
        for col, img in enumerate(sub_imgs):
            nimgs[pad + row * (h + pad):pad + row * (h + pad) + h,
            pad + col * (w + pad):pad + col * (w + pad) + w, ] = img
    return nimgs

vr = decord.VideoReader('v_HulaHoop_g07_c04.avi',width=32,height=32)
plt.figure(dpi=100)
plt.imshow(make_grid_from_nplist(vr[::4].asnumpy(),ncol=8))


import torch

from lumo.core import Params
from lumo.core.metaclasses import make_dicts, make_dict
from lumo.exp import SimpleExperiment
from .factory import OptimFactory, InterpFactory


class TrainerExperiment(SimpleExperiment):

    @property
    def log_dir(self):
        return self.test_root

    @property
    def params_fn(self):
        res = self.test_file('params.json')
        return res

    @property
    def board_args(self):
        key = 'tensorboard_args'
        if self.has_prop(key):
            return self.get_prop(key)
        else:
            log_dir = self.test_dir('board')
            res = {
                'filename_suffix': '.bd',
                'log_dir': log_dir,
            }
            self.dump_info('tensorboard_args', res)
            return res

    @property
    def saver_dir(self):
        res = self.blob_dir('saver')
        return res

    def dump_train_info(self, epoch: int):
        self.dump_info('trainer', {
            'epoch': epoch
        }, append=True)


class ReimplementExperiment(TrainerExperiment):
    pass


class TrainerPropVar(type):
    def __new__(cls, name, bases, attrs: dict, **kwds):
        for base in bases:
            for key, value in base.__dict__.items():  # type:(str,Any)
                if key.endswith("__"):
                    continue
                if isinstance(value, set):
                    v = attrs.setdefault(key, set())
                    v.update(value)
                elif isinstance(value, dict):
                    v = attrs.setdefault(key, dict())
                    v.update(value)

        clazz = type.__new__(cls, name, bases, dict(attrs))

        make_dicts(clazz, [
            '_prop',
            '_cmp',
            '_rev_index',
            '_call_order',
        ])

        make_dict(clazz, '_state_dicts', {
            'optims': set(),
            'models': set(),
            'others': set(),
            'tensor.th': set(),
            'tensor.np': set(),
        })
        return clazz


class TrainerParams(Params):
    OPTIM = OptimFactory
    SCHE = INTERP = InterpFactory

    def __init__(self):
        super().__init__()
        self.epoch = 10
        self.device = "cuda:0" if torch.cuda.is_available() else "cpu"



from lumo.contrib.optim import lr_scheduler
from lumo import attr


def test_scheduler_is_attr():
    a = attr()
    a.sche = lr_scheduler.CosScheduler(start=0, end=1, left=0, right=1)
    assert a.sche.get(0) == a.sche(0) == 0

    b = attr.from_dict(a.jsonify())
    assert b.sche.__class__.__name__ == lr_scheduler.CosScheduler.__name__
    assert b.sche.get(0) == b.sche(0) == 0
    assert a.hash() == b.hash()



"""

cd {{repo_root}}
python -m lumo.data supervised

python -m lumo.data selfsupervised

python -m lumo.data semisupervised
from datasets.objdect

python -m lumo.data https://github.com/example/dataset objdect


from datasets.objdect
"""



import sys

from absl import app
from tensorboard import default
from tensorboard import main_lib
from tensorboard import program
from tensorboard.plugins import base_plugin
from tensorboard.uploader import uploader_subcommand
from tensorboard.util import tb_logging

logger = tb_logging.get_logger()

from typing import overload


@overload
def run_board(logdir, host=None, port=None): pass


def run_board(logdir, **kwargs):
    main_lib.global_init()

    tensorboard = program.TensorBoard(
        plugins=default.get_plugins(),
        subcommands=[uploader_subcommand.UploaderSubcommand()],
    )
    try:
        app.run(tensorboard.main,
                argv=[
                    sys.argv[0],
                    '--logdir', logdir,
                ],
                flags_parser=tensorboard.configure)
    except base_plugin.FlagsError as e:
        print("Error: %s" % e, file=sys.stderr)
        sys.exit(1)



from abc import ABC
from typing import Dict, Any, Optional


class RecordBackend(ABC):
    def __init__(self, location, *args, **kwargs):
        self.location = location

    def log(self, data: Dict[str, Any],
            step: Optional[int] = None,
            commit: Optional[bool] = None,
            sync: Optional[bool] = None):
        raise NotImplementedError()

    def log_image(self, image_array, caption=None):
        raise NotImplementedError()
    def log_audio(self, image_array, caption=None):
        raise NotImplementedError()
    def log_image(self, image_array, caption=None):
        raise NotImplementedError()



"""
Used for recording data
"""
from collections import OrderedDict
from collections.abc import ItemsView
from numbers import Number
from typing import Union, Iterator, Tuple, Mapping, Sequence

import numpy as np
import torch

from lumo.utils.fmt import to_ndarray, detach, is_scalar
from lumo.core import PropVar


class Meter(metaclass=PropVar):
    def __init__(self):
        self._rec = {}
        self._avg = {}

    def sorted(self) -> 'Meter':
        m = Meter()
        m._rec = self._rec
        m._avg = OrderedDict()
        m._prop = self._prop
        for k in sorted(self._avg.keys()):
            m._avg[k] = self._avg[k]
        return m

    def todict(self):
        return self._rec

    @property
    def _stage(self):
        return self._prop.get('stage', 'default')

    @_stage.setter
    def _stage(self, value):
        self._prop['stage'] = value

    def __setattr__(self, key: str, value):
        if key.startswith('_'):
            super(Meter, self).__setattr__(key, value)
        else:
            self[key] = value

    def __getattr__(self, item):
        return self[item]

    def __getitem__(self, item):
        return self._rec[item]

    def __setitem__(self, key, value):
        value = to_ndarray(value)

        stg = self._avg.get(key, None)
        isscalar = value.size == 1

        if stg is None:
            dtype = value.dtype.name

            if self._stage in {'min', 'max'} and not isscalar:
                raise ValueError(
                    f'Only support min/max(a) operator on scalar metrics, but got data of shape {value.shape}.')
            elif self._stage in {'min', 'max', 'sum', 'mean', 'smean'} and 'str' in dtype:
                raise ValueError(f'Only support min/max/sum/mean operator on tensor metrics, but got type {dtype}.')

            if self._stage == 'default':
                if isscalar:
                    if 'int' in dtype:
                        self._stage = 'last'
                    else:
                        self._stage = 'mean'
                else:
                    self._stage = 'last'

            self._avg[key] = self._stage

        if isscalar:
            value = value.item()

        self._rec[key] = value
        self._stage = 'default'

    def __repr__(self):
        return ' | '.join([f'{k}: {v}' for k, v in self._rec.items()])

    def __iter__(self):
        yield from self.keys()

    @property
    def sum(self):
        self._stage = 'sum'
        return self

    @property
    def mean(self):
        self._stage = 'mean'
        return self

    @property
    def last(self):
        self._stage = 'last'
        return self

    @property
    def max(self):
        self._stage = 'max'
        return self

    @property
    def min(self):
        self._stage = 'min'
        return self

    @property
    def smean(self):
        self._stage = 'smean'
        return self

    def update(self, dic: Mapping) -> 'Meter':
        for k, v in dic.items():
            self[str(k)] = v
        return self

    def serialize(self) -> OrderedDict:
        res = OrderedDict()
        for k, v in self.items():
            res[k] = f'{v}'
        return res

    def items(self) -> ItemsView:
        return self._rec.items()

    def keys(self):
        return self._rec.keys()

    @staticmethod
    def from_dict(dic: Mapping):
        m = Meter()
        for k, v in dic.items():
            m[k] = v
        return m

    def scalar_items(self) -> Iterator[Tuple[str, Number]]:
        for k, v in self.items():
            nd = to_ndarray(v)
            if is_scalar(nd):
                yield k, nd.item()


class AvgItem:
    SLIDE_WINDOW_SIZE = 100
    EXP_WEIGHT = 0.75

    def __init__(self, item, gb_method):
        item_ = detach(item)
        self.gb_method = gb_method  # groupby method
        self.acc = [item_]
        self.c = 1
        self.cur = item
        self.last = item_
        self.offset = item_
        self.nd = to_ndarray(item)
        self.isint = 'int' in self.nd.dtype.name
        self.isnumber = (self.isint or 'float' in self.nd.dtype.name) and isinstance(item, (np.ndarray,
                                                                                            torch.Tensor))
        self.isscalar = self.nd.size == 1
        if not self.isscalar and gb_method in {'min', 'max'}:
            raise AssertionError(f'{gb_method} method only support scaler')

    def __repr__(self):
        """
        simpler but more time-comsuming method could be some math function, not in if-else branch, like
            prec =  max(min(8, int(np.ceil(np.log10((1 / (self.offset + 1e-10)))))), 1)
            fmt_str = f'{{:.{prec}f}}'
            return fmt_str.format(res)
        """
        res = self.res
        if self.isscalar:
            res = to_ndarray(res).item()
            if self.isint:
                return f"{res}"
            elif self.isnumber:
                # return f'{res:.4f}'
                if self.offset < 1e-8:
                    return f'{res:.10f}'
                elif self.offset < 1e-6:
                    return f'{res:.8f}'
                elif self.offset < 1e-4:
                    return f'{res:.6f}'
                return f'{res:.4f}'
            else:
                return f'{res}'
        else:
            return f'{res}'

    __str__ = __repr__

    def update(self, item):
        self.cur = item
        item = detach(item)

        avg = self.gb_method
        if self.isnumber:
            self.offset = self.offset * AvgItem.EXP_WEIGHT + abs(item - self.last) * (1 - AvgItem.EXP_WEIGHT)

        if avg == 'slide':
            self.acc.append(item)
            if len(self.acc) > AvgItem.SLIDE_WINDOW_SIZE:
                self.acc.pop(0)
            self.last = self.cur
        elif avg in {'mean', 'sum'}:
            self.acc[0] = self.acc[0] + item
            self.c += 1
        elif avg == 'max':
            self.last = max(self.cur, self.last)
        elif avg == 'min':
            self.last = min(self.cur, self.last)
        elif avg == 'last':
            self.last = item

    @property
    def res(self):
        avg = self.gb_method
        if avg == 'slide':
            return np.mean(self.acc)
        if avg == 'mean':
            return self.acc[0] / self.c
        elif avg == 'sum':
            return self.acc[0]
        elif avg in {'max', 'min', 'last'}:
            return self.last
        return self.cur



import random
from lumo.base_classes.list import llist


class MemoryBank:
    def __init__(self, queue_size=512, lis_cls=llist, *args, **kwargs):
        self.queue_size = queue_size
        assert self.queue_size > 0, 'queue size must be larger than 0.'
        self._lis = lis_cls()

    def __len__(self):
        return len(self._lis)

    def __iter__(self):
        return iter(self._lis)

    def __getitem__(self, item):
        return self._lis[item]

    def isempty(self):
        return len(self._lis) == 0

    def isfull(self):
        return len(self._lis) == self.queue_size

    def pop(self, index=0, default=None):
        if self.isempty() or index >= len(self):
            return default
        return self._lis.pop(index)

    def top(self):
        return self._lis[0]

    def topk(self, k=1):
        return self._lis[:k]

    def tail(self):
        return self._lis[-1]

    def tailk(self, k=1):
        return self._lis[-k:]

    def push(self, item):
        self._lis.append(item)
        if self.isfull():
            self.pop()

    def choice(self, default=None):
        if self.isempty():
            return default
        return random.choice(self._lis)

    def apply(self, func):
        self._lis = self._lis.__class__(func(i) for i in self._lis)



import os
import random
import sys
import time
import traceback
from pathlib import Path
from typing import Union
from lumo.utils.logger import Logger
from lumo.decorators.process import call_on_main_process_wrap
from lumo.proc.path import blobroot, exproot, cache_dir, libhome
from .base import ExpHook
from ..proc.dist import is_dist, is_main, local_rank
from ..proc.path import exproot, local_dir
from ..utils import safe_io as io
from ..utils.fmt import can_be_filename


def checkdir(path: Union[Path, str]):
    if isinstance(path, str):
        os.makedirs(path, exist_ok=True)
    elif isinstance(path, Path):
        path.mkdir(parents=True, exist_ok=True)
    return path


class Experiment:
    """
    (by default), the directory structure is as following:
    .lumo (libroot)
        - experiments # (exp_root) record information (e.g., .log, params files, etc.)
            - {experiment-name-1}
                - {test-1}
                - {test-2}
            - {experiment-name-2}
                - {test-1}
                - {test-2}
        - blob # (blob_root) record big binary files (like model state dicts)
            - {experiment-name-1}
                - {test-1}
                - {test-2}
            - {experiment-name-2}
                - {test-1}
                - {test-2}
        - metric # (metric_root) record metrics (by trainer.database)
            - {experiment-name-1}
                - {test-1}
                - {test-2}
            - {experiment-name-2}
                - {test-1}
                - {test-2}
    """

    def __init__(self, exp_name: str, root=None):
        if not can_be_filename(exp_name):
            raise ValueError(f'Experiment name should be a ligal filename(bettor only contain letter or underline),'
                             f'but got {exp_name}.')

        self._prop = {}
        self._prop['exp_name'] = exp_name
        self._hooks = {}
        if root is None:
            root = libhome()
        self._root = Path(os.path.abspath(root))
        self.add_exit_hook(self.end)
        self.logger = Logger()

    @property
    def exp_name(self):
        return self._prop['exp_name']

    @property
    def _test_name(self):
        return self._prop.get('test_name', None)

    @_test_name.setter
    def _test_name(self, value):
        self._prop['test_name'] = value

    @property
    def test_name_with_dist(self):
        """
        Create different test_name for each process in multiprocess training.
        Returns: in main process, will just return test_name itself,
        in subprocess,  "{test_name}.{local_rank()}"
        """
        raw = self.test_name
        if not is_dist():
            return raw

        if is_main():
            return raw
        else:
            return f"{raw}.{local_rank()}"

    @property
    def test_name(self):
        """Assign unique space(directory) for this test"""
        if self._test_name is None:
            if is_dist():  # if train in distribute mode, subprocess will wait a few seconds to wait main process.
                flag_fn = f'.{os.getppid()}'
                if is_main():
                    self._test_name = self._create_test_name()
                    fn = self.exp_file(flag_fn)
                    with open(fn, 'w') as w:
                        w.write(self._test_name)
                else:
                    time.sleep(random.randint(2, 4))
                    fn = self.exp_file(flag_fn)
                    if os.path.exists(fn):
                        with open(fn, 'r') as r:
                            self._test_name = r.readline().strip()
            else:
                self._test_name = self._create_test_name()

        return self._test_name

    def _create_test_name(self):
        from lumo.proc.date import timehash
        from ..utils.fmt import strftime
        fs = os.listdir(self.exp_root)
        date_str = strftime('%y%m%d')
        fs = [i for i in fs if i.startswith(date_str)]
        _test_name = f"{date_str}.{len(fs):03d}.{timehash()[-6:-4]}t"
        return _test_name

    @property
    def root_branch(self):
        val = self._root
        return checkdir(val)

    @property
    def lib_root(self):
        return self.root_branch.as_posix()

    @property
    def exp_branch(self):
        val = Path(exproot()).joinpath(self.exp_name)
        return checkdir(val)

    @property
    def blob_branch(self):
        val = Path(blobroot()).joinpath(self.exp_name, self.test_name)
        return checkdir(val)

    @property
    def test_branch(self):
        val = self.exp_branch.joinpath(self.test_name)
        return checkdir(val)

    def dump_info(self, key: str, info: dict, append=False, info_dir='info', set_prop=True):
        fn = self.test_file(f'{key}.json', info_dir)
        if append:
            old_info = self.load_info(key, info_dir=info_dir)
            old_info.update(info)
            info = old_info
        if set_prop:
            self.set_prop(key, info)
        io.dump_json(info, fn)

    def load_info(self, key: str, info_dir='info'):
        fn = self.test_file(f'{key}.json', info_dir)
        if not os.path.exists(fn):
            return {}
        return io.load_json(fn)

    def dump_string(self, key: str, info: str):
        fn = self.test_file(f'{key}.str', 'text')
        io.dump_text(info, fn)
        self.set_prop(key, info)

    def load_string(self, key: str):
        fn = self.test_file(f'{key}.str', 'text')
        if not os.path.exists(fn):
            return ''
        return io.load_text(fn)

    @property
    def tags(self):
        tags = {}
        for path in self.test_branch.joinpath('tags').glob('tag.*.json'):
            ptags = io.load_json(path.as_posix())  # type: dict
            tags.setdefault(path.suffixes[0].strip('.'), []).extend(ptags.keys())
        return tags

    def add_tag(self, tag: str, name_space: str = 'default'):
        self.dump_info(f'tag.{name_space}', {
            tag: None
        }, append=True, info_dir='tags', set_prop=False)

    def exp_file(self, filename, *args):
        """

        Args:
            filename:
            *args:
            mkdir:

        Returns:

        """
        parent = self.exp_branch.joinpath(*args)
        return checkdir(parent).joinpath(filename).as_posix()

    def test_file(self, filename, *args):
        parent = self.test_branch.joinpath(*args)
        return checkdir(parent).joinpath(filename).as_posix()

    def exp_dir(self, *args):
        """

        Args:
            filename:
            *args:
            mkdir:

        Returns:

        """
        parent = self.exp_branch.joinpath(*args)
        return checkdir(parent).as_posix()

    def root_file(self, filename, *args):
        parent = self.root_branch.joinpath(*args)
        return checkdir(parent).joinpath(filename).as_posix()

    def root_dir(self, *args):
        """

        Args:
            filename:
            *args:
            mkdir:

        Returns:

        """
        parent = self.root_branch.joinpath(*args)
        return checkdir(parent).as_posix()

    def test_dir(self, *args):
        parent = self.test_branch.joinpath(*args)
        return checkdir(parent).as_posix()

    def blob_file(self, filename, *args):
        parent = self.blob_branch.joinpath(*args)
        return checkdir(parent).joinpath(filename).as_posix()

    def blob_dir(self, *args):
        """

        Args:
            filename:
            *args:
            mkdir:

        Returns:

        """
        parent = self.blob_branch.joinpath(*args)
        return checkdir(parent).as_posix()

    def __enter__(self):
        self.start()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        extra = {}
        if exc_type is not None:
            exc_type = traceback.format_exception_only(exc_type, exc_val)[-1].strip()
            extra['exc_type'] = exc_type
            extra['end_info'] = str(exc_type)
            extra['end_code'] = 1
            extra['exc_stack'] = "".join(traceback.format_exception(exc_type, exc_val, exc_tb))
        self.end(**extra)

    @call_on_main_process_wrap
    def add_exit_hook(self, func):
        import atexit
        def exp_func():
            func(self)

        atexit.register(exp_func)

    @call_on_main_process_wrap
    def initial(self):
        self.add_tag(self.__class__.__name__, 'exp_type')
        self.dump_info('execute', {
            'repo': self.project_root,
            'cwd': os.getcwd(),
            'exec_file': sys.argv[0],
            'exec_bin': sys.executable,
            'exec_argv': sys.argv
        })

    @call_on_main_process_wrap
    def start(self):
        if self.get_prop('start', False):
            return
        self.initial()
        self.set_prop('start', True)
        for hook in self._hooks.values():  # type: ExpHook
            hook.on_start(self)
        return self

    @call_on_main_process_wrap
    def end(self, end_code=0, *args, **extra):
        if not self.get_prop('start', False):
            return
        if self.get_prop('end', False):
            return
        self.set_prop('end', True)
        for hook in self._hooks.values():  # type: ExpHook
            hook.on_end(self, end_code=end_code, *args, **extra)
        return self

    @property
    def repo_name(self):
        """repository name"""
        return self.project_name

    @property
    def project_name(self):
        """same as repository name, directory name of project root"""
        return os.path.basename(self.project_root)

    @property
    def project_root(self):
        return local_dir()

    @property
    def exp_root(self):
        """path to multiple tests of this experiment"""
        return self.exp_branch.as_posix()

    @property
    def test_root(self):
        """path to record information of one experiment"""
        return self.test_branch.as_posix()

    @property
    def blob_root(self):
        """path to storing big binary files"""
        return self.blob_branch.as_posix()

    def __getitem__(self, item):
        return self._prop[item]

    def __setitem__(self, key, value):
        self._prop[key] = value

    def get_prop(self, key, default=None):
        return self._prop.get(key, default)

    def has_prop(self, key):
        return key in self._prop

    def set_prop(self, key, value):
        self._prop[key] = value

    @property
    def properties(self):
        return self._prop

    @property
    def paths(self) -> dict:
        return {
            'root': self.root_branch.as_posix(),
            'exp_root': self.exp_root,
            'test_root': self.test_root,
            'blob_root': self.blob_root,
        }

    @property
    def enable_properties(self) -> set:
        return set(self._prop.keys())

    @call_on_main_process_wrap
    def set_hook(self, hook: ExpHook):
        hook.regist(self)
        self.logger.info(f'Register {hook}.')
        self._hooks[hook.__class__.__name__] = hook
        self.add_tag(hook.__class__.__name__, 'hooks')
        return self

    def load_prop(self):
        for f in os.listdir(self.test_dir('info')):
            key = os.path.splitext(f)[0]
            self.set_prop(key, self.load_info(key))

        for f in os.listdir(self.test_dir('text')):
            key = os.path.splitext(f)[0]
            self.set_prop(key, self.load_string(key))

    @classmethod
    def from_disk(cls, path):
        from .finder import is_test_root
        if not is_test_root(path):
            raise ValueError(f'{path} is not a valid test_root')

        test_root = Path(path)
        root = test_root.parent.parent.parent.as_posix()
        self = cls(test_root.parent.name, root=root)
        self._test_name = test_root.name
        self.load_prop()
        return self

    @property
    def exec_argv(self):
        execute_info = self.get_prop('execute')
        try:
            return [os.path.basename(execute_info['exec_bin']), *execute_info['exec_argv']]
        except:
            return []


class SimpleExperiment(Experiment):

    def __init__(self, exp_name: str, root=None):
        super().__init__(exp_name, root)
        from . import exphook
        self.set_hook(exphook.LastCmd())
        self.set_hook(exphook.LockFile())
        self.set_hook(exphook.GitCommit())
        self.set_hook(exphook.RecordAbort())
        self.set_hook(exphook.Diary())
        self.set_hook(exphook.TimeMonitor())
        self.set_hook(exphook.FinalReport())



from accelerate.accelerator import Accelerator as _Accelerator
from accelerate.data_loader import prepare_data_loader


class Accelerator(_Accelerator):
    def prepare_data_loader(self, data_loader):
        return prepare_data_loader(
            data_loader,
            None,
            num_processes=self.num_processes,
            process_index=self.process_index,
            split_batches=self.split_batches,
            put_on_device=self.device_placement,
            rng_types=self.rng_types.copy(),
            dispatch_batches=self.dispatch_batches,
        )



import fire
import os
from .functional import *
from ..exp import Experiment

doc = """
Usage:
# create templete directory 
lumo init [dir]

# easier way to open tensorboard 
# lumo board [--logdir=<logdir>]
# lumo board [--test=<test_name>] # find test_name and tensorboard it 
# lumo board  # default open ./board

# lumo mark <test_name>

# restore code snapshot of some test
lumo reset <test_name>

# archive code snapshot of some test
lumo archive <test_name>

# print log file
lumo log <test_name>

# print params of this test
lumo params <test_name>

# <test_name>/--test=<test_name>/--test_name=<test_name>

# TODO
lumo config local --k=v
lumo config global --k=v

# get a free port
lumo port

"""


class Main:
    def sum(self, tid):
        """

        Args:
            tid: test_name or test_root

        Returns:

        """
        from ..exp.finder import summary_experiment
        summary_experiment(tid)

    def today(self):
        pass

    def init(self, path):
        git_init(path)
        print(os.path.abspath(path))

    def extract(self, test_root, output=None, verbose=True):
        exp = Experiment.from_disk(test_root)
        test_extract(test_root, output=output, verbose=verbose)

    def clone(self, arg: str, alias: str = None):
        """
        if template:
            git clone template_map[arg] alias
        else:
            git clone <arg> alias

        Args:
            arg: url or template name
            template: template id
            alias: alias

        Returns:

        """
        if '/' not in arg:
            _, path = git_clone_from_template(arg, alias)
        else:
            _, path = git_clone(arg, alias)
        git_init(path)


fire.Fire(Main())
exit(0)



from typing import NoReturn, Union, overload, Optional

from torch.utils.data import DataLoader

from lumo.core import PropVar, ParamsType, TrainStage
from .loader import DataLoaderType
from .loader import DataLoaderSide


class DataModule(metaclass=PropVar):
    def __init__(self, params: ParamsType = None):
        self.params = params

    @staticmethod
    def _parse_dataset(loader):
        if isinstance(loader, DataLoader):
            return loader.dataset
        elif isinstance(loader, DataLoaderSide):
            return loader.dataset
        return None

    @property
    def train_dataset(self):
        return self._parse_dataset(self.train_dataloader)

    @property
    def test_dataset(self):
        return self._parse_dataset(self.test_dataloader)

    @property
    def val_dataset(self):
        return self._parse_dataset(self.val_dataloader)

    @property
    def train_dataloader(self) -> Optional[DataLoaderType]:
        return self.get_loader_with_stage(TrainStage.train)

    @property
    def test_dataloader(self) -> Optional[DataLoaderType]:
        return self.get_loader_with_stage(TrainStage.test)

    @property
    def val_dataloader(self) -> Union[NoReturn, DataLoaderType]:
        return self.get_loader_with_stage(TrainStage.val)

    def get_loader_with_stage(self, stage: TrainStage) -> DataLoaderType:
        res = self._prop.get(stage.value, None)
        if res is None:
            self.idataloader(self.params, stage)
            res = self._prop.get(stage.value, None)
        return res

    def __getitem__(self, key):
        return self._prop.get(key, None)

    @overload
    def regist_dataloader(self, train=None, test=None, val=None):
        ...

    def regist_dataloader(self, **kwargs: dict):
        for k, v in kwargs.items():
            self._prop[k] = v

    def regist_dataloader_with_stage(self, stage: TrainStage, dl: DataLoaderType):
        self._prop[stage.value] = dl

    def idataloader(self, params: ParamsType = None, stage: TrainStage = None):
        pass



from typing import Iterable


def recursive_get(dic: dict, key: str, default=None):
    for key in key.split('.'):
        if isinstance(dic, dict):
            dic = dic.get(key, default)
        else:
            return dic

    return dic


def flatten_dict(df, key: str, keys: Iterable, prefix: str = None, default=None):
    if prefix is None:
        prefix = key
    for k in keys:
        df[f'{prefix}.{k}'] = df[key].apply(lambda x: recursive_get(x, k, default))


def flatten_params(df, *keys: str):
    return flatten_dict(df, 'params', keys, prefix='p')


def flatten_metric(df, *keys: str):
    return flatten_dict(df, 'metric', keys, prefix='m')



import json
import sys
import textwrap
from pprint import pformat
from typing import Any, List, NewType

import fire
from joblib import hash
from pathlib import Path
from omegaconf import DictConfig, OmegaConf, DictKeyType
from omegaconf._utils import _ensure_container

from .raises import BoundCheckError, NewParamWarning

# arange_param = namedtuple('arange_param', ['default', 'left', 'right'], defaults=[None, float('-inf'), float('inf')])
# choice_param = namedtuple('choice_param', ['default', 'choices'], defaults=[None, []])

__all__ = ['BaseParams', 'Params', 'ParamsType']


class Arange:
    def __init__(self, default=None, left=float('inf'), right=float('inf')):
        self.default = default
        self.left = left
        self.right = right

    def __repr__(self):
        return f"Arange: {self.default}, [{self.left}, {self.right}] "


class Choices:
    def __init__(self, default=None, choices=None):
        if choices is None:
            choices = []
        self.default = default
        self.choices = choices

    def __repr__(self):
        return f"Choice: [{self.default}], {self.choices}"


def _get_item(dic, keys: List[str]):
    if len(keys) == 1:
        return DictConfig.__getitem__(dic, keys[0])
    else:
        nex = DictConfig.__getitem__(dic, keys[0])
        if isinstance(nex, (dict, DictConfig)):
            return _get_item(nex, keys[1:])
        else:
            raise KeyError(keys)


def _set_item(dic, keys: List[str], value):
    if len(keys) == 1:
        if isinstance(value, dict):
            value = dic(value)
        DictConfig.__setitem__(dic, keys[0], value)
    else:
        try:
            nex = _get_item(dic, keys[:1])
        except KeyError:
            nex = DictConfig({})
            DictConfig.__setitem__(dic, keys[0], nex)
        _set_item(nex, keys[1:], value)


def _safe_repr(values: Any) -> str:
    return pformat(values)


def _padding_mod(st: str, offset=7, mod=4):
    """
    123 \\
    1   \\
    12312341    \\
    1231
    Args:
        strs:
        mod:

    Returns:

    """
    size = len(st)
    if size < offset:
        return st.ljust(offset, ' ')

    mnum = mod - len(st) % mod
    # if mnum == 0:
    #     mnum = mod
    return st.ljust(size + mnum, ' ')


def safe_param_repr(values: List[tuple], level=1) -> str:
    """

    Args:
        values:
        level:

    Returns:

    """
    res = [(f"{k}={_safe_repr(v)},", anno) for k, v, anno in values]

    # res = textwrap.fill('\n'.join(res))
    res = '\n'.join([_padding_mod(i, offset=16, mod=4) + f'  # {anno}' for i, anno in res])

    return textwrap.indent(res, '    ')


class BaseParams(DictConfig):
    def __init__(self):
        super().__init__({}, flags={'no_deepcopy_set_nodes': True})
        # self._set_flag('no_deepcopy_set_nodes', True)
        self.__dict__["_prop"] = {}

    def __setattr__(self, key: str, value: Any) -> None:
        if key != '_prop':
            # if isinstance(value, BaseParams):
            #     self._prop.setdefault('key_type', {})[key] = type(value)

            if isinstance(value, (Arange, Choices)):
                res = self._prop.get('constrain', {})
                res[key] = value
                self._prop['constrain'] = res
                value = value.default

            if key in self._prop.setdefault('constrain', {}):
                self._check(key, value)

        super().__setattr__(key, value)

    def __setitem__(self, key: DictKeyType, value: Any) -> None:
        if key != '_prop':
            # if isinstance(value, BaseParams):
            #     self._prop.setdefault('key_type', {})[key] = type(value)

            if isinstance(value, (Arange, Choices)):
                self._prop.setdefault('constrain', {})[key] = value
                value = value.default

            if key in self._prop.setdefault('constrain', {}):
                self._check(key, value)

        super().__setitem__(key, value)

    def __getattr__(self, key: str) -> Any:
        res = super().__getattr__(key)
        # key_type = self._prop.setdefault('key_type', {}).get(key, None)
        # if key_type is not None:
        #     res = key_type.from_kwargs(**res)
        return res

    def _check(self, name, value):
        bound = self._prop['constrain'][name]
        if isinstance(bound, Arange) and not (bound.left <= value and value <= bound.right):
            raise BoundCheckError(
                f"value of param '{name}' should in range [{bound.left}, {bound.right}], but got {value}")
        elif isinstance(bound, Choices) and value not in bound.choices:
            raise BoundCheckError(f"value of param '{name}' should in values {bound.choices}, but got {value}")

    def __getitem__(self, key: DictKeyType) -> Any:
        return super().__getitem__(key)

    def __repr__(self):
        def _arg_to_str(k, v):
            res = self._prop.get('constrain', {}).get(k, None)
            if res is not None:
                return f'{res}, {type(v).__name__}'

            key_type = self._prop.get('key_type', {}).get(k, None)
            if key_type is not None:
                return f'{key_type.__name__}'

            return f'{type(v).__name__}'

        args = [(k, v) for k, v in self.items()]
        args = [(k, v, _arg_to_str(k, v)) for k, v in args]

        args_str = safe_param_repr(args)

        return "{}.Space".format(self.__class__.__name__) + '(\n' + args_str + '\n)'

    def arange(self, default, left=float("-inf"), right=float("inf")) -> Arange:
        """
        Make sure some value is into some range.

        Examples:
            params.batch_size = params.arange(20,10,100)
            print(params.batch_size) # will print '20' as default.
            params.batch_size = 300 # will raise an Exception
            params.batch_size = 50
            print(params.batch_size) # will print 50

        Args:
            k: key of the value
            default: default value
            left: left interval
            right: right interval

        Returns:
            arange_param(default, left, right)
        """
        if left < default and default < right:
            return Arange(default, left, right)
        else:
            raise BoundCheckError(f"value {default}' should in range [{left}, {right}].")

    def choice(self, *choices) -> Choices:
        """
        Make sure some value is into some limited values.

        Examples:
            params.dataset = params.choice('cifar10','cifar100')
            print(params.dataset) # will print 'cifar10' as default.
            params.dataset = 'mnist' # will raise an Exception
            params.dataset = 'cifar100'
            print(params.dataset) # will print 'cifar100'

        Args:
            k: key of the value
            *choices: value can be used for key

        Returns:
            choice_param(choices[0], choices)


        """
        return Choices(choices[0], choices)

    def from_dict(self, dic: dict):
        for k, v in dic.items():
            self[k] = v
        return self

    def from_kwargs(self, **kwargs):
        return self.from_dict(kwargs)

    def from_json(self, file):
        self.update(json.loads(Path(file).read_text()))
        return self

    def from_yaml(self, file):
        self.update(OmegaConf.load(file))
        return self

    def from_args(self, argv: list = None):
        if argv is None:
            argv = sys.argv

        def func(**kwargs):
            if '_help' in kwargs:
                print(self)
                exit()
                return

            if '_json' in kwargs:
                self.from_json(kwargs['_json'])
                return

            for k, v in kwargs.items():
                # try:
                #     _get_item(self, k.split('.'))
                # except:
                # self[k] = v
                _set_item(self, k.split('.'), v)

        fire.Fire(func)
        return self

    def from_hydra(self, config_path, config_name):
        import hydra
        hydra.compose()

        @hydra.main(config_path=config_path, config_name=config_name)
        def inner(cfg):
            return cfg

        self.update(inner())
        return self

    def to_dict(self):
        cfg = _ensure_container(self)
        container = OmegaConf.to_container(cfg, resolve=False, enum_to_str=True)
        return container

    def to_json(self, file=None):
        info = json.dumps(self.to_dict(), ensure_ascii=False, indent=2)
        if file is None:
            return info
        return Path(file).write_text(info, encoding='utf-8')

    def to_yaml(self, file=None):
        info = OmegaConf.to_yaml(self)
        if file is None:
            return info
        return Path(file).write_text(info, encoding='utf-8')

    @classmethod
    def Space(cls, **kwargs):
        return cls().from_dict(kwargs)

    def __hash__(self):
        return int(self.hash(), 16)

    def hash(self) -> str:
        return hash(str(self))

    def iparams(self):
        pass

    @classmethod
    def init_from_kwargs(cls, **kwargs):
        return cls().from_dict(kwargs)


class Params(BaseParams):
    pass


ParamsType = NewType('ParamsType', Params)



"""
Help you find the experiments you have done.
"""

# from git import Commit, Repo
#
#
# class TestProp():
#     @property
#     def command(self):
#         raise NotImplementedError()
#
#     @property
#     def exp_name(self):
#         raise NotImplementedError()
#
#     @property
#     def exp_root(self):
#         raise NotImplementedError()
#
#     @property
#     def project_name(self):
#         raise NotImplementedError()
#
#     @property
#     def repo_root(self):
#         raise NotImplementedError()
#
#     @property
#     def repo_hash(self):
#         raise NotImplementedError()
#
#     @property
#     def start_time(self):
#         raise NotImplementedError()
#
#     @property
#     def end_time(self):
#         raise NotImplementedError()
#
#     @property
#     def end_code(self):
#         raise NotImplementedError()
#
#     @property
#     def grepo(self) -> Repo:
#         raise NotImplementedError()
#
#     @property
#     def gcommit(self) -> Commit:
#         raise NotImplementedError()
#
#     @property
#     def commit_hash(self) -> str:
#         raise NotImplementedError()
#
#     @property
#     @lru_cache(1)
#     def lines(self) -> Dict[str, Dict]:
#         raise NotImplementedError()
#
#     @property
#     @lru_cache(1)
#     def jsons(self):
#         raise NotImplementedError()
#
#     @property
#     def pkl_keys(self):
#         raise NotImplementedError()
#
#     @property
#     def uuid(self):
#         raise NotImplementedError()
#
#     @property
#     def largest_epoch(self):
#         raise NotImplementedError()
#
#
# @dataclass(unsafe_hash=True)
# class Test(TestProp):
#     tid: str
#     root: str
#
#     @property
#     def command(self):
#         exe = self.jsons[EXP.EXECUTE]['exec_bin']
#         argv = self.jsons[EXP.EXECUTE]['exec_argv']
#         return ' '.join([exe, *argv])
#
#     @property
#     def exp_name(self):
#         return os.path.basename(self.exp_root)
#
#     @property
#     def exp_root(self):
#         return os.path.dirname(self.root)
#
#     @property
#     def project_name(self):
#         return os.path.basename(self.repo_root)
#
#     @property
#     def repo_root(self):
#         return self.jsons[EXP.PROJECT]['root']
#
#     @property
#     def repo_hash(self):
#         return self.jsons[EXP.PROJECT]['hash']
#
#     @property
#     def start_time(self):
#         val = self.jsons[EXP.STATE].get('start', None)
#         if val is not None:
#             val = date_from_str(val)
#         return val
#
#     @property
#     def end_time(self):
#         val = self.jsons[EXP.STATE].get('end', None)
#         if val is not None:
#             val = date_from_str(val)
#         return val
#
#     @property
#     def end_code(self):
#         return self.jsons[EXP.STATE].get('end_code', 1)
#
#     @property
#     def grepo(self) -> Repo:
#         raise NotImplementedError()
#         # repo = load_repo(self.repo_root)
#         # return repo
#
#     @property
#     def gcommit(self) -> Commit:
#         from gitdb.util import hex_to_bin
#         return Commit(self.grepo, hex_to_bin(self.commit_hash))
#
#     @property
#     def commit_hash(self) -> str:
#         """None if nocommit in this test, or commit hash will be returned"""
#         return self.jsons.get(EXP.GIT, {}).get('commit', None)
#
#     @property
#     def lines(self) -> Dict[str, str]:
#         dir_ = os.path.join(self.root, FN.D_LINE)
#         if not os.path.exists(dir_):
#             return attr()
#         fs_ = os.listdir(dir_)
#         fs_ = [i for i in fs_ if i.endswith(FN.SUFFIX.D_LINE)]
#         res = attr()
#         for f in fs_:
#             k = os.path.splitext(f)[0]
#             v = io.load_string(os.path.join(dir_, f))
#             if v is None:
#                 os.remove(os.path.join(dir_, f))
#                 continue
#             res[k] = v
#         return res
#
#     @property
#     def jsons(self):
#         dir_ = os.path.join(self.root, FN.D_JSON)
#         if not os.path.exists(dir_):
#             return attr()
#         fs_ = os.listdir(dir_)
#         fs_ = [i for i in fs_ if i.endswith('.json')]
#         res = attr()
#         for f in fs_:
#             k = os.path.splitext(f)[0]
#             v = io.load_json(os.path.join(dir_, f))
#             if v is None:
#                 continue
#             res[k] = v
#         return res
#
#     @property
#     @lru_cache(1)
#     def pkl_keys(self):
#         dir_ = os.path.join(self.root, FN.D_JSON)
#         if not os.path.exists(dir_):
#             return set()
#
#         fs_ = os.listdir(dir_)
#         fs_ = [i for i in fs_ if i.endswith('.pkl')]
#         res = set()
#         for f in fs_:
#             k = os.path.splitext(f)[0]
#             res.add(k)
#         return res
#
#     @property
#     def uuid(self):
#         return self.lines.get(EXP.UUID, '')
#
#     @property
#     def largest_epoch(self):
#         return self.jsons.get(EXP.TRAINER, {}).get('epoch', 0)
#
#     def pkl(self, key):
#         """
#
#         Args:
#             key:
#
#         Returns:
#
#         """
#         dir_ = os.path.join(self.root, FN.D_JSON)
#         f = os.path.join(dir_, f"{key}.pkl")
#         return io.load_state_dict(f)
#
#
# class Condition():
#     def filter(self, test: Test) -> bool:
#         raise NotImplementedError()
#
#     def __call__(self, tests: List[Test]):
#         return [i for i in tests if self.filter(i)]
#
#
# class LambdaCondition(Condition):
#     def __init__(self, func):
#         self.func = func
#
#     def filter(self, test: Test) -> bool:
#         return self.func(test)
#
#
# class Identity(Condition):
#
#     def filter(self, test: Test) -> bool:
#         return True
#
#
# class Or(Condition):
#     def __init__(self, *conditions: Condition):
#         self.conditions = list(conditions)
#
#     def add(self, condition: Condition):
#         self.conditions.append(condition)
#
#     def filter(self, test: Test) -> bool:
#         return any([i.filter(test) for i in self.conditions])
#
#
# class And(Or):
#     def filter(self, test: Test) -> bool:
#         return all([i.filter(test) for i in self.conditions])
#
#
# class Query:
#     def __init__(self):
#         self.conditions = And()
#
#     def success(self):
#         self.conditions.add(LambdaCondition(lambda x: x.end_code == 0))
#         return self
#
#     def failed(self):
#         self.conditions.add(LambdaCondition(lambda x: x.end_code != 0))
#         return self
#
#     def has_info(self, key):
#         self.conditions.add(LambdaCondition(lambda x: x.jsons.get(key, None) is not None))
#         return self
#
#     def has_line(self, key):
#         self.conditions.add(LambdaCondition(lambda x: x.lines.get(key, None) is not None))
#         return self
#
#     def in_time(self, start=None, end=None):
#         if start is not None:
#             self.conditions.add(LambdaCondition(lambda x: x.start_time > start is not None))
#         if end is not None:
#             self.conditions.add(LambdaCondition(lambda x: x.end_time > end is not None))
#         return self
#
#     def train_longer_than(self, epoch=0):
#         self.conditions.add(LambdaCondition(lambda x: x.largest_epoch > epoch))
#         return self
#
#     def train_shorter_than(self, epoch=1):
#         self.conditions.add(LambdaCondition(lambda x: x.largest_epoch <= epoch))
#         return self
#
#     def in_repo(self, name=None, path=None, full=False):
#         if name is not None:
#             if full:
#                 self.conditions.add(LambdaCondition(lambda x: name == x.project_name))  # type:Test
#             else:
#                 self.conditions.add(LambdaCondition(lambda x: name in x.project_name))  # type:Test
#         elif path is not None:
#             self.conditions.add(LambdaCondition(lambda x: compare_path(path, x.repo_root)))  # type:Test
#         else:
#             path = git_dir(path)
#             if path is not None:
#                 return self.in_repo(path=path)
#             else:
#                 return Identity()
#
#         return self
#
#     def build(self):
#         return self.conditions
#
#
# class Finder:
#     def __init__(self):
#         fn = os.path.join(libhome(), FN.REPOSJS)
#         if os.path.exists(fn):
#             res = io.load_json(fn)
#         else:
#             res = {}
#         self.meta = res
#         self._test_dirs = []
#         self._tests = Tests()
#         self.refresh()
#
#     def refresh(self):
#         fn = os.path.join(libhome(), FN.TESTLOG)
#         if not os.path.exists(fn):
#             return
#         res = io.load_string(fn)
#         if res is None:
#             res = ''
#         res = res.split('\n')
#
#         self._test_dirs = list(OrderedDict.fromkeys(
#             filter(lambda x: os.path.exists(x.strip()), res)
#         ).keys())
#         self._test_dirs = sorted(self._test_dirs, key=lambda x: os.stat(x).st_atime)
#
#         self._tests = Tests(Test(os.path.basename(i), i) for i in self._test_dirs)
#
#     def tests(self, *conditions, nonflag=None):
#         """can be found in each expdir"""
#         tests = self._tests
#         for condition in conditions:
#             tests = condition(tests)
#         if len(tests) == 0:
#             return nonflag
#         return Tests(tests)
#
#     def lastest(self) -> Test:
#         return self._tests[-1]
#
#     def n_lastest(self, n) -> List[Test]:
#         return self._tests[-n:]
#
#
# class Tests(TestProp, List[Test]):
#
#     def _to_df(self, value: list, column: str):
#         import pandas as pd
#         df = pd.DataFrame()
#         df[column] = value
#         df.index = self.names
#         return df
#
#     @property
#     def command(self):
#         return self._to_df([i.command for i in self], 'command')
#
#     @property
#     def exp_name(self):
#         return self._to_df([i.exp_name for i in self], 'exp_name')
#
#     @property
#     def exp_root(self):
#         return self._to_df([i.exp_root for i in self], 'exp_root')
#
#     @property
#     def project_name(self):
#         return self._to_df([i.project_name for i in self], 'project_name')
#
#     @property
#     def repo_root(self):
#         return self._to_df([i.repo_root for i in self], 'repo_root')
#
#     @property
#     def repo_hash(self):
#         return self._to_df([i.repo_hash for i in self], 'repo_hash')
#
#     @property
#     def start_time(self):
#         return self._to_df([i.start_time for i in self], 'start_time')
#
#     @property
#     def end_time(self):
#         return self._to_df([i.end_time for i in self], 'end_time')
#
#     @property
#     def end_code(self):
#         return self._to_df([i.end_code for i in self], 'end_code')
#
#     @property
#     def grepo(self):
#         return self._to_df([i.grepo for i in self], 'grepo')
#
#     @property
#     def gcommit(self):
#         return self._to_df([i.gcommit for i in self], 'gcommit')
#
#     @property
#     def commit_hash(self):
#         return self._to_df([i.commit_hash for i in self], 'commit_hash')
#
#     @property
#     def lines(self):
#         return [i.lines for i in self]
#
#     @property
#     def jsons(self):
#         return [i.jsons for i in self]
#
#     @property
#     def pkl_keys(self):
#         return [i.pkl_keys for i in self]
#
#     @property
#     def names(self):
#         return [i.tid for i in self]
#
#     @property
#     def uuid(self):
#         return self._to_df([i.uuid for i in self], 'uuid')
#
#     @property
#     def latest(self):
#         return self[-1]
#
#     @property
#     def largest_epoch(self):
#         return self._to_df([i.largest_epoch for i in self], 'largest_epoch')
#
#     @property
#     def root(self):
#         return self._to_df([i.root for i in self], 'root')
#
#     def delete(self):
#         for i in self:
#             shutil.rmtree(i.root)
#
#
# F = Finder()
# Q = Query()



import os
import json

GLOBAL_DEFAULT = {
    'home': os.path.expanduser("~/.lumo/"),
    'cache_dir': os.path.expanduser("~/.cache/lumo"),
    'exp_root': os.path.expanduser("~/.lumo/experiments"),
    'blob_root': os.path.expanduser("~/.lumo/blob"),
    'metric_root': os.path.expanduser("~/.lumo/metric"),
}


def global_config_path():
    return os.path.expanduser("~/.lumorc.json")


def local_config_path():
    from lumo.utils.repository import git_dir
    res = git_dir()
    if res:
        return os.path.join(res, ".lumorc.json")
    return None


def get_config(path, default):
    if path is None:
        return default

    if os.path.exists(path):
        try:
            with open(path, encoding='utf-8') as r:
                config = json.load(r)
            return config
        except Exception as e:
            print(f'Error read {path}')

    os.makedirs(os.path.dirname(path), exist_ok=True)
    with open(path, 'w', encoding='utf-8') as w:
        json.dump(default, w)
    return default


def create_runtime_config():
    glob_cfg = get_config(global_config_path(), GLOBAL_DEFAULT)
    local_cfg = get_config(local_config_path(), {})
    cfg = GLOBAL_DEFAULT
    cfg.update(glob_cfg)
    cfg.update(local_cfg)
    return cfg


def pretain_model_path(name):
    pass


glob = create_runtime_config()

HOME = glob.get("home")
CACHE_ROOT = glob.get('cache_dir')
EXP_ROOT = glob.get('exp_root')
BLOB_ROOT = glob.get('blob_root')
METRIC_ROOT = glob.get('metric_root')



import re
from datetime import datetime
from setuptools import setup, find_packages


def extract_version():
    return re.search(
        r'__version__ = "([\d.d\-]+)"',
        open('src/lumo/__init__.py', 'r', encoding='utf-8').read()).group(1)


if __name__ == '__main__':
    setup(
        name='lumo',
        version=extract_version(),
        description='library to manage your pytorch experiments.',
        long_description_content_type='text/markdown',
        url='https://github.com/pytorch-lumo/lumo',
        author='sailist',
        author_email='sailist@outlook.com',
        license_files=('LICENSE',),
        include_package_data=True,
        classifiers=[
            'Programming Language :: Python :: 3.6',
            'Programming Language :: Python :: 3.7',
            'Programming Language :: Python :: 3.8',
            'Programming Language :: Python :: 3.9',
        ],
        package_dir={"": "src"},
        keywords='lumo',
        packages=find_packages('src'),
        entry_points={
            'console_scripts': [
                'lumo = lumo.cli.cli:main'
            ]
        },
    )



from pprint import pformat
import os
from lumo.proc.path import libhome
from lumo.utils.fmt import indent_print

from . import Experiment


def find_experiments():
    return os.listdir(os.path.join(libhome(), 'experiment'))


def list_test_names_from_experiment(experiment):
    return os.listdir(os.path.join(libhome(), 'experiment', experiment))


def find_path_from_test_name(test_name: str):
    if not is_test_name(test_name):
        return None

    date, _, _ = test_name.split('.')
    f = os.path.join(libhome(), 'diary', f'{date}.log')
    if os.path.exists(f):
        with open(f) as r:
            for line in r:
                time, path = line.split(', ')
                if test_name in path:
                    return path.strip()
    return None


def is_test_name(test_name: str):
    chunk = test_name.split('.')
    if len(test_name) != 14:
        return False

    if len(chunk) != 3:
        return False
    for k, c in zip(chunk, [6, 3, 3]):
        if len(k) != c:
            return False
    for k in chunk[:2]:
        try:
            int(k)
        except:
            return False
    return True


def is_test_root(test_root: str):
    test_name = os.path.basename(test_root.rstrip('/'))
    return is_test_name(test_name)


def format_experiment(exp: Experiment):
    return {
        'Properties': exp.properties,
        'tags': exp.tags,
        'paths': exp.paths,
        'exec_argv': exp.exec_argv,
    }


def get_experiment_name(test_root:str):
    return os.path.basename(os.path.dirname(test_root.rstrip('/')))

def get_test_name(test_root:str):
    return os.path.basename(test_root.rstrip('/'))

def ensure_test_root(tid: str):
    if is_test_name(tid):
        test_root = find_path_from_test_name(tid)
        if test_root is None:
            return None
    elif is_test_root(tid):
        test_root = tid
    else:
        return None

    return test_root


def summary_experiment(tid: str):
    test_root = ensure_test_root(tid)
    if test_root is None:
        return test_root

    exp = Experiment.from_disk(test_root)

    print('Properties:')
    indent_print(pformat(exp.properties))
    print('Tags:')
    indent_print(pformat(exp.tags))
    print('Use paths:')
    indent_print(pformat(exp.paths))
    print('Execute:')
    indent_print(' '.join(exp.exec_argv))
    print('-----------------------------------')



from typing import Callable

from lumo.proc.dist import is_main


def call_on_main_process_wrap(func) -> Callable:
    def inner(*args, **kwargs):
        if is_main():
            return func(*args, **kwargs)

    return inner



from abc import ABC, ABCMeta
from collections import OrderedDict

__all__ = ['PropVar', 'OrderedPropVar', 'AttrPropVar', 'ABCPropVar']

from typing import List
from functools import wraps


def make_dicts(type_name, names: List[str], dic_type=dict):
    def outer(func):
        @wraps(func)
        def inner(cls, *args, **kwargs):
            self = func(cls)
            for name in names:
                object.__setattr__(self, name, dic_type())
                # setattr(self, name, dic_type())
            return self

        return inner

    type_name.__new__ = outer(type_name.__new__)


def make_dict(type_name, name: str, default):
    def outer(func):
        @wraps(func)
        def inner(cls, *args, **kwargs):
            self = func(cls)
            setattr(self, name, default)
            return self

        return inner

    type_name.__new__ = outer(type_name.__new__)


class PropVar(type):
    """
    """

    def __new__(cls, name, bases, attrs: dict, **kwds):
        clazz = type.__new__(cls, name, bases, dict(attrs))
        make_dicts(clazz, ['_prop'])
        return clazz


class OrderedPropVar(type):
    """
    """

    def __new__(cls, name, bases, attrs: dict, **kwds):
        clazz = type.__new__(cls, name, bases, dict(attrs))
        make_dicts(clazz, ['_prop'], OrderedDict)
        return clazz


class AttrPropVar(type):
    """
    """

    def __new__(cls, name, bases, attrs: dict, **kwds):
        from .attr import Attr
        clazz = type.__new__(cls, name, bases, dict(attrs))
        make_dicts(clazz, ['_prop'], Attr)
        return clazz


class ABCPropVar(ABCMeta):

    def __new__(cls, name, bases, attrs: dict, **kwds):
        from .attr import Attr
        clazz = type.__new__(cls, name, bases, dict(attrs))
        make_dicts(clazz, ['_prop', '_content'], Attr)
        return clazz


class Merge(type):
    """
    元类，用于将子类和父类共有字典，集合时，子类的覆盖行为改为合并父类的字典，集合

    由于用途特殊，仅识别类变量中以下划线开头的变量
    ::
        class A(metaclass=Merge):
            _dicts = {"1": 2, "3": 4}

        class B(A):
            _dicts = {"5":6,7:8}

        print(B._dicts)

    result:
    >>> {'5': 6, '3': 4, '1': 2, 7: 8}
    """

    def __new__(cls, name, bases, attrs: dict, **kwds):
        for base in bases:
            for key, value in base.__dict__.items():  # type:(str,Any)
                if key.endswith("__"):
                    continue
                if isinstance(value, set):
                    v = attrs.setdefault(key, set())
                    v.update(value)
                elif isinstance(value, dict):
                    v = attrs.setdefault(key, dict())
                    v.update(value)

        return type.__new__(cls, name, bases, dict(attrs))



import inspect
import os


def analyse_module_dependency(module, mem=None, root=None):
    if mem is None:
        mem = {}

    if inspect.ismodule(module):
        try:
            fn = getattr(module, '__file__', None)
        except RuntimeError:
            return mem
        if fn is None or fn in mem:
            return mem
        if os.path.commonprefix([os.path.abspath(fn), root]) != root:
            return mem
        mem[fn] = module
        for k, v in module.__dict__.items():
            analyse_module_dependency(v, mem, root)
    else:
        try:
            module = inspect.getmodule(module)
            if module is not None:
                analyse_module_dependency(module, mem, root)
        except:
            pass

    return mem



import time

import os
import random
from joblib import hash
import mmap
from lumo.utils.exithook import wrap_before


class Lock:
    def __init__(self, name, sleep=1, group=None):
        from lumo.proc.path import cache_dir
        if group is None:
            group = ""

        self.size = 10000
        self.flag = f'{os.getpid()}'
        self.flagsize = len(self.flag)
        self.pos = int(hash(name), 16) % (self.size - len(self.flag))
        self.file = os.path.join(cache_dir(), f'LUMO_LOCK_V3_{group}')
        if not os.path.exists(self.file):
            with open(self.file, 'w') as w:
                w.write('0' * self.size)

        self.sleep = sleep
        self.fhdl = None
        wrap_before(self.clear)

    def clear(self, *_, **__):
        self.release()

    def abtain(self):
        mulp = 0
        self.fhdl = r = open(self.file, 'r+b')
        mm = mmap.mmap(r.fileno(), 0, access=mmap.ACCESS_WRITE)
        flag = f'{os.getpid()}'
        while True:
            mulp += 1
            if mulp > 10:
                raise TimeoutError(f'Can not abtain resource of {self.file}')

            if mm[self.pos:self.pos + len(flag)].decode() != '0' * len(flag):
                time.sleep(random.randint(mulp, mulp ** 2))
                continue

            mm[self.pos:self.pos + len(flag)] = flag.encode()
            mm.flush()

            if mm[self.pos:self.pos + len(flag)].decode() != flag:
                mm.close()
                time.sleep(random.randint(mulp, mulp ** 2))
                continue

            return True

    def release(self):
        if self.fhdl is None:
            return

        r = self.fhdl
        mm = mmap.mmap(r.fileno(), 0, access=mmap.ACCESS_WRITE)
        flag = b'0' * self.flagsize
        mm[self.pos:self.pos + len(flag)] = flag
        mm.flush()
        r.close()
        self.fhdl = None
        return True

    def __enter__(self):
        self.abtain()

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.release()



from tensorboard.backend.event_processing import event_accumulator
from typing import List, Optional, Dict
from collections import defaultdict


class TBReader:
    def __init__(self, root: str, name: Optional[str] = None):
        self.event = event_accumulator.EventAccumulator(root,
                                                        size_guidance={  # see below regarding this argument
                                                            event_accumulator.COMPRESSED_HISTOGRAMS: 500,
                                                            event_accumulator.IMAGES: 4,
                                                            event_accumulator.AUDIO: 4,
                                                            event_accumulator.SCALARS: 0,
                                                            event_accumulator.HISTOGRAMS: 1,
                                                        })

        self.root = root
        self.name = name

    def __enter__(self):
        self.event.Reload()

    def scalars(self) -> List[str]:
        with self:
            return self.event.Tags()['scalars']


class GroupTBReader:
    def __init__(self, roots: List[str], names: Optional[List[str]] = None):
        if names is None:
            names = [f'Tz{i}' for i in range(len(roots))]
        assert len(roots) == len(names)
        self.readers = [TBReader(i) for i in roots]
        self.names = names

    def scalars(self) -> Dict[str, List[str]]:
        res = defaultdict(list)
        for r, n in zip(self.readers, self.names):
            for tag in r.scalars():
                res[tag].append(n)
        return res

    def get_table_of(self, scalar_tag):
        """"""
        pass



""""""
import warnings

from torch import nn
import torch
from copy import deepcopy
from typing import TypeVar

Module = TypeVar('Module', bound=nn.Module)


def empty_train(*args, **kwargs):
    return args[0]


import numpy as np

np.random.rand()


def EMA(model: Module, alpha=0.999, force_eval=False) -> Module:
    """
    EMA(model: Module, alpha=0.999, force_eval=False)

        An Exponential Moving Average(EMA) wrapper for nn.Module

        Examples:
        --------
        >>> model = ...
        >>> ema_model = EMA(model,alpha=0.999)

            ...

            ema_model.step()
            # or ema_model.step(0.99)

        Args:
            model:
            alpha:
            force_eval:

        Returns:

    """
    ema_model = deepcopy(model)

    [i.requires_grad_(False) for i in ema_model.parameters()]
    param_keys = set([k for k, _ in ema_model.named_parameters()])
    buffer_keys = set([k for k, _ in ema_model.named_buffers()])  # for Norm layers

    def step(alpha_=None):

        if alpha_ is None:
            alpha_ = alpha

        with torch.no_grad():
            for (k, ema_param), (_, param) in zip(ema_model.state_dict().items(), model.state_dict().items()):
                if k in param_keys:
                    ema_param.data.copy_(alpha_ * ema_param + (1 - alpha_) * param)
                elif k in buffer_keys:
                    ema_param.data.copy_(param)

    forward_ = ema_model.forward

    def forward(*args, **kwargs):
        with torch.no_grad():
            return forward_(*args, **kwargs)

    ema_model.forward = forward

    if hasattr(ema_model, 'step'):
        warnings.warn(f'EMA use `step` function to update module paramters, '
                      f'the defined `step` function in class {model.__class__.__name__} will be replaced.')

    ema_model.step = step
    ema_model.eval()
    if force_eval:
        ema_model.train = empty_train
    return ema_model



"""
A hack function to monitor GPU memory usage, then occupy its access.
"""
import functools
import os
import subprocess
import time
from functools import partial

import torch

from lumo.core.tree import tree
from lumo.utils import re

match_mem = re.compile('([0-9]+) +([0-9]+)[^|]* ([0-9]+)MiB')


class DeviceMem:
    def __init__(self):
        self.line_mem = tree()

    def _parse_device_pid_mem_pair(self, lines):
        for lid, line in enumerate(lines):
            res = re.search(match_mem, line)
            if res is not None:
                _device, _pid, _mib = [int(i) for i in res.groups()]
                self.line_mem[_device][_pid] = lid
                yield _device, _pid, _mib

    def try_parse(self, lines, pid, device):
        """ try parse mem from cached lid directly.
        Returns:
             -1 means failed.
             others means successd and its memory.

        """
        lid = self.line_mem[device][pid]
        if isinstance(lid, dict):
            return -1
        elif lid > len(lines):
            return -1
        else:
            res = re.search(match_mem, lines[lid])
            if res is None:
                return -1
            else:
                _device, _pid, _mib = [int(i) for i in res.groups()]
                if _pid == pid and _device == device:
                    return _mib
                else:
                    return -1

    def re_parse(self, lines, pid, device):
        _res = self.try_parse(lines, pid, device)
        if _res != -1:
            return _res

        for _device, _pid, _mib in self._parse_device_pid_mem_pair(lines):
            if _pid == pid and _device == device:
                return _mib

        return 0

    def _get_nvidia_smi(self):
        proc = subprocess.Popen(['nvidia-smi'], stdout=subprocess.PIPE)
        lines = proc.stdout.readlines()
        return [i.decode() for i in lines]

    def _device_equal(self, da, db):
        if isinstance(da, (int, str)):
            da = torch.device(da)
        if isinstance(db, (int, str)):
            db = torch.device(db)
        return da == db

    def get_device_release_mem(self, device):
        """ get device memory left."""
        s_pid = os.getpid()
        total = self.get_device_mem(device)
        for _device, _pid, _mib in self._parse_device_pid_mem_pair(self._get_nvidia_smi()):
            if self._device_equal(device, _device):
                total -= _mib

        return total

    def get_device_mem(self, device):
        """  returns device total memory(unit: MB)        """
        return torch.cuda.get_device_properties(device).total_memory // (1024 * 1024)

    def get_pid_device_mem(self, pid, device):
        """
        尽可能有效率的得到进程在某设备下占用的显存（通过命令行程序调用获取）
        :param pid:
        :param device:
        :return:
        """
        if isinstance(device, torch.device):
            device = device.index
        elif isinstance(device, str):
            device = torch.device(device).index

        lines = self._get_nvidia_smi()
        mib = self.try_parse(lines, pid, device)
        if mib == -1:
            mib = self.re_parse(lines, pid, device)

        return mib


_memer = DeviceMem()
_pid = os.getpid()

if torch.cuda.is_available():
    get_pid_device_mem = partial(_memer.get_pid_device_mem, pid=_pid, device=torch.cuda.current_device())


class memory(object):
    r"""
    优雅的抢卡
    Args:
        memory: 需要占用的内存，以 MB 为单位
        device: 需要占用内存的设备
        hold:
        unit:
    Example::
        >>> import lumo
        >>> with lumo.memory(5000):
        ...   y = x * 2

        >>> @lumo.memory(1024)
        ... def doubler(x):
        ...     ...

        >>> lumo.memory(10000).start()
        ... # do something

    Why use nvidia-smi to get memory useage? see:
        https://github.com/pytorch/pytorch/issues/12873
    """

    def __init__(self, memory, device=None, hold=False) -> None:
        super().__init__()
        if device is None:
            device = torch.cuda.current_device()
        if isinstance(device, (str, int)):
            device = torch.device(device)

        self.need = memory
        self.device = device
        self.hold = hold
        self.exc_time = 0
        self.acc = 5
        self.mem = []
        self.last_success = _memer.get_pid_device_mem(_pid, self.device)

    def copy(self, pid: int, wait: bool = True):
        self.need = _memer.get_pid_device_mem(pid, self.device)
        if wait:
            self.wait(pid)
        else:
            self.start()

    def wait(self, pid):
        while _memer.get_pid_device_mem(pid, self.device) > 0:
            time.sleep(0.5)
        self.start()

    def immediately(self, pre_init=False):
        """
        等待，直到内存有空间后，开始申请相应显存，优雅，礼貌，推荐
        Args:
            pre_init: 是否初始化 CUDA（这将在一开始消耗一定显存），默认为 False，即不抢占任何内存，
                直到设备释放足够空间后开始抢占。
        """
        while True:
            _left = _memer.get_device_release_mem(self.device)
            _allocated = _memer.get_pid_device_mem(_pid, self.device)

            if pre_init and _allocated == 0:
                self._malloc(1, init=True)
                continue

            _need = self.need - _allocated
            if _need < 0:
                return self.end()

            if _left > _need:
                if _allocated == 0:
                    self._malloc(1, init=True)
                    continue

                res = self._malloc(_need)
                time.sleep(0.5)
                if res:
                    return self.end()

            print("need {}Mb, {}Mb allocated, "
                  "waiting for {}Mb released, "
                  "but only {}Mb left.".format(self.need,
                                               _allocated, _need,
                                               _left), end='\r')

    def _malloc(self, size, init=False):
        """ unit: mb """
        try:
            tmp = torch.rand(size, 1048576 // 4, device=self.device)
            if not init:
                self.mem.append(tmp)
            return True
        except Exception as e:
            return False

    def end(self):
        print()
        if self.hold:
            print('press keyboardinterrupt to end')
            try:
                while True:
                    # do some Fake thing
                    self.mem[-1].random_()
                    time.sleep(0.1)
            except KeyboardInterrupt:
                print('continue')

    def start(self, immediately=True):
        if immediately:
            self.immediately()
        else:
            self.invade()

    def invade(self, unit=5):
        """一点一点的侵占，有多少占用多少，直到申请满为止，比较粗鲁，不友好，不推荐"""
        try:
            while self.last_success < self.need:
                res = self._malloc(unit + self.acc)
                if res:
                    self.acc += unit
                    self.last_success = _memer.get_pid_device_mem(_pid, self.device)
                    time.sleep(0.1)
                else:
                    self.exc_time += 1
                    self.acc = max(self.acc - unit, 0)
                    time.sleep(0.5)
                print('{}/{}Mb, try={}, pid={}'.format(self.last_success,
                                                       self.need,
                                                       self.exc_time,
                                                       os.getpid()), end='\r')
            self.end()
        except KeyboardInterrupt:
            print('\nabort.')

    def __enter__(self):
        self.invade()

    def __exit__(self, *args):
        del self.mem[:]
        torch.cuda.empty_cache()
        return True

    def __call__(self, func):
        @functools.wraps(func)
        def decorate_no_grad(*args, **kwargs):
            with self:
                return func(*args, **kwargs)

        return decorate_no_grad

    @staticmethod
    def hold_current():
        count = torch.cuda.device_count()
        mems = [_memer.get_pid_device_mem(_pid, i) for i in range(count)]
        for i, mem in enumerate(mems):
            if mem > 0:
                memory(mem, device=i, hold=(i == count - 1)).start()



from accelerate.utils import recursively_apply
import torch


def send_to_device(tensor, device, non_blocking=False):
    """
    Recursively sends the elements in a nested list/tuple/dictionary of tensors to a given device.

    Args:
        tensor (nested list/tuple/dictionary of `torch.Tensor`):
            The data to send to a given device.
        device (`torch.device`):
            The device to send the data to.

    Returns:
        The same data structure as `tensor` with all tensors sent to the proper device.
    """
    if not isinstance(device, torch.device):
        device = torch.device(device)

    if 'mps' in device.type:
        non_blocking = False

    def _send_to_device(t, device):
        return t.to(device, non_blocking=non_blocking)

    def _has_to_method(t):
        return hasattr(t, "to")

    return recursively_apply(_send_to_device, tensor, device, test_type=_has_to_method)
